"""Web app component that provides RSS feeds on IMAP mailboxes, and
simple message reading.

TODO:

Copyright (C) 2002 l.m.orchard <deus_x@pobox.com>
http://www.decafbad.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details, published at 
http://www.gnu.org/copyleft/gpl.html
"""

import os
import string
import Caffinate

import re
import imaplib
import rfc822
import StringIO

from xml.sax import saxutils

### TODO: How can I match for '&nbsp;' as whitespace / termination of URL
url_protocols = "http|https|ftp|news|nntp|mailto|telnet|gopher"
url_pattern   = "((?:(?:%s):[^\s\]\"\>\)]+))" % (url_protocols)
url_re = re.compile(url_pattern)

class ImapRss(Caffinate.ApplicationComponent):

	def __init__(self, conf_fn):
		Caffinate.ApplicationComponent.__init__(self, conf_fn)

	def imap_connect(self, request):
		host = self.get_config('imap_host')
		user = self.get_config('imap_user')
		pswd = self.get_config('imap_pass')
		mail_prefix   = self.get_config('path_prefix')
		mail_path_sep = self.get_config('path_separator', 'global', '/')
		mailbox_path  = self.get_config \
						('mailbox_path',
						 request.get_parameter('mailbox', 'global'),
						 "")

		imap = imaplib.IMAP4(host)
		imap.login(user, pswd)
				
		if mailbox_path == "":
			imap.select()
		else:
			mailbox_full_name = "%s%s%s" % \
								(mail_prefix, mail_path_sep, mailbox_path)
			imap.select(mailbox_full_name)

		return imap

	def _emit_link(self, mo):
		return "<a href=\"%s\">%s</a>" % (mo.group(0),mo.group(0))

	def _nbsps_to_spaces(self, mo):
		return string.replace(mo.group(0), "&nbsp;", " ")

	def _htmlize_body_line(self, line):
		### TODO: How can I match for '&nbsp;' as termination of URL
		out = line
		out = url_re.sub(self._emit_link, out)
		out = string.replace(out, " ", "&nbsp;")
		out = re.sub("(\<[^\>]+\>)", self._nbsps_to_spaces, out)
		return out

	def web_read(self, app, request, response):
		imap = self.imap_connect(request)
		mailbox = request.get_parameter('mailbox', 'global')

		msg_num = request.get_parameter('msg', '1')

		typ, data = imap.fetch(msg_num, '(RFC822)')
		mesg = rfc822.Message(StringIO.StringIO(data[0][1]), 0)

		body_text = string.join(map(self._htmlize_body_line, mesg.fp.readlines() ), "<br>\n")

		ns = { 'header' : mesg,
			   'body'   : body_text }

		imap.logout()

		response.set_template('imap-read.html.tmpl')
	   	response.add_namespace(ns)

	def web_rss(self, app, request, response):
		imap = self.imap_connect(request)
		mailbox = request.get_parameter('mailbox', 'global')
		mailbox_path  = self.get_config('mailbox_path', mailbox, "")
		#skip_word = self.get_config('skipword', mailbox)

		imap = self.imap_connect(request)

		num_msgs     = request.get_parameter('msgs', 15)
		chars_in_msg = request.get_parameter('text_len', 512)
				
		typ, search_list = imap.search(None, 'ALL')
		search_items = string.split(search_list[0])
		search_items.reverse()
			
		items_ns_list = []
		ns = { 'title'       : 'Mailbox: %s' % mailbox_path,
			   'link'        : 'None',
			   'description' : 'This is mailbox "%s"' % mailbox_path,
			   'msgs'        : [] }
			
 		for msg_num in search_items[0:num_msgs]:
				
			typ, data = imap.fetch(msg_num, "(BODY.PEEK[HEADER])")
			#header = rfc822.Message(msg(data[0][1]), 0)
			header = rfc822.Message(StringIO.StringIO(data[0][1]), 0)
		
			try:             mail_from = string.join(header.getaddr('from'), " ")
			except KeyError: mail_from = '(sender unknown)'
				
			try:             mail_date = header['date']
			except KeyError: mail_date = '(date unknown)'
				
			try:             mail_subject = header['subject']
			except KeyError: mail_subject = '(no subject)'
				
			typ, data = imap.fetch(msg_num, "(BODY.PEEK[1]<0.%s>)" % chars_in_msg)
			mail_text = string.join(map(lambda x: string.strip(x),
										string.split(data[0][1], '\015\012')))
				
			item_ns = {'title'        : mail_subject,
					   'link'         : "%s?method=read&mailbox=%s&msg=%s" %
					   (request.get_base_url(), mailbox_path, msg_num),
					   'description'  : "(From: %s) %s" % (mail_from, mail_text) }
			
			for key in item_ns.keys():
				item_ns[key] = saxutils.escape(item_ns[key])
					
			ns['msgs'].append(item_ns)

		imap.logout()

		response.set_template('imap-rss.tmpl')
	   	response.add_namespace(ns)
